/* RealSpeak.h

RealSpeak(TM) driver.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Peter Dikant <peter.dikant@mobilexag.de>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include "RealSpeak.h"

/**
 * the default constructor will create the basic tts instances and also store a pointer
 * to the logging subsystem.
 * @param log a pointer to the logging object
 */
RealSpeak::RealSpeak(Logger *log, const char* engine)
    :TextToSpeech(log)
{
    log->log("<debug> Initializing RealSpeak Engine");

    strcpy(enginePath, engine);
    ttsParm.nLanguage = TTS_LANG_GERMAN;
    ttsParm.nOutputType = TTS_LINEAR_16BIT;
    ttsParm.nFrequency = TTS_FREQ_8KHZ;
    ttsParm.nVoice = TTS_RS_VOICE_FEMALE;
    ttsParm.nInputDataType = TTS_DATA_TYPE_TEXT;
    ttsParm.nOutputDataType = TTS_DATA_TYPE_PCM;
    ttsParm.cbFuncs.TtsSourceCb = CbTtsSource;
    ttsParm.cbFuncs.TtsDestCb = CbTtsDestination;
    ttsParm.cbFuncs.TtsEventCb = CbTtsEventNotify;
    ttsParm.cbFuncs.numCallbacks = 3;
    ttsParm.szLibLocation = enginePath;

    int ret = TtsInitialize(&hInst, NULL, &ttsParm, (void *)this);
    if (ret != 0) {
	log->logf("<error> Error initializing RealSpeak Engine: %d", ret);
    }
    log->log("<debug> RealSpeak Engine initialized.");

    output = (unsigned char*)malloc(TTS_OUTPUT_BUFFER + 44);
    memcpy(output, wavHeader, 44);
    InitializeCriticalSection(&criticalSection);
}

TTSRETVAL RealSpeak::CbTtsEventNotify (void *pAppData, void *ppBuffer, U16 nDataSize, U16 event) {
    RealSpeak *me = (RealSpeak*)pAppData;
    //me->log->logf(CLOG_LEVEL_NOTICE, "Received RealSpeak Event: %d", event);
    return 0;
}

TTSRETVAL RealSpeak::CbTtsSource(void *lpAppData, void *lpDataBuffer, U32 nBufferSize, U32 *lpDataSize) {
    RealSpeak *me = (RealSpeak*)lpAppData;
    //me->log->logf(CLOG_LEVEL_DEBUG, "TTS: Callback Source called.");
    
    if (strlen(me->text) == 0) {
	//me->log->log(CLOG_LEVEL_DEBUG, "TTS: sending TTS_ENDOFDATA");
	*lpDataSize = 0;
	return TTS_ENDOFDATA;
    }
    if (nBufferSize >= strlen(me->text)) {
	memcpy(lpDataBuffer, me->text, strlen(me->text));
	*lpDataSize = strlen(me->text);
	strcpy(me->text, "");
    } else {
	memcpy(lpDataBuffer, me->text, nBufferSize);
    	*lpDataSize = nBufferSize;
	strcpy(me->text, &me->text[nBufferSize]);
    }
    //me->log->log(CLOG_LEVEL_DEBUG, "TTS: sending TTS_SUCCESS");
    return TTS_SUCCESS;
}

void* RealSpeak::CbTtsDestination(void *lpAppData, U16 nDatatype, void *lpData, U32 nDataSize, U32 *pnBufferSize) {
    RealSpeak *me = (RealSpeak*)lpAppData;
    //me->log->logf(CLOG_LEVEL_DEBUG, "TTS: Callbeck Destination called.");
    if (nDataSize == 0) {
	// first callback, report size of buffer and buffer address
	*pnBufferSize = TTS_OUTPUT_BUFFER;
	return (void *)&me->output[me->outputSize + 44];
    } else {
	//me->log->logf(CLOG_LEVEL_DEBUG, "generated %d bytes data", nDataSize);
	me->outputSize += (long)nDataSize;
	*pnBufferSize = TTS_OUTPUT_BUFFER;
	return (void *) &me->output[nDataSize + 44];
    }
}

/**
 * free used ressources
 */
RealSpeak::~RealSpeak()
{
	log->log("<debug> Deleting RealSpeak Engine");
	int ret = TtsUninitialize(hInst);
	if (ret != 0) {
	    log->logf("<error> Error closing RealSpeak Engine: %d", ret);
	}
	free(output);
	DeleteCriticalSection(&criticalSection);
}

/*
 * load a User Dictionary
 */
void RealSpeak::loadDictionary(const char* dictionary)
{
	int ret;
	if(dictionary) {
	    ret = TtsLoadUsrDict(0, &userdict, (char*)dictionary);
	    if(ret != 0) {
		log->logf("<error> Couldn't load dictionary \"%s\"", dictionary);
	    } else {
		ret = TtsEnableUsrDict(hInst, userdict);
		if(ret != 0) {
		    log->logf("<error> Couldn't enable dictionary \"%s\"", dictionary);
		} else {
		    log->logf("<notice> Dictionary \"%s\" loaded", dictionary);
		}
	    }
	}
}

/**
 * set implementation specific parameters
 */
void RealSpeak::setParams(const char *key, const char *value)
{
	/* test which parameters are for us */
	if(!strncmp(key, "realspeak.", 10))
	{
	    const char*key2 = &key[10];
	    if(!strcmp(key2, "dictionary"))
	    {
		loadDictionary(value);
	    }
	}
}

/**
 * this method will convert a given text to speech and output this over the microsoft
 * audio mapper, thus creating direct audio output on the soundcard.
 * @param text the text to be converted to speech
 */
void RealSpeak::text2Audio(const char *text)
{
	log->log("<verbose> text2Audio is not implemented in RealSpeak");
}

/**
 * create speech directly to a wave file
 * @param text the text to be spoken
 * @param filenam name and path of the file to be created
 */
void RealSpeak::text2File(const char *text, const char *filename)
{
    EnterCriticalSection(&criticalSection);
    strncpy(this->text, text, TTS_TEXT_BUFFER);

    // start conversion
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputSize);

    unsigned int wavelen = outputSize + 36;
    memcpy(output + 4, &wavelen, 4);
    wavelen = outputSize;
    memcpy(output + 40, &wavelen, 4);

    FILE *fp = fopen(filename, "wb");
    fwrite(output, sizeof(unsigned char), outputSize + 44, fp);
    fclose(fp);
    LeaveCriticalSection(&criticalSection);
}


/**
 * this method will create speech that is stored in a memory buffer. the created speech has the format pcm mono with 
 * 8kHz sampling rate and 16 bit. The memory structure can be directly used for input to the capi
 * subsystem.
 * @param text this text will be spoken
 * @return oubuffer pointer to the output buffer (the memory will be allocated)
 * @return size the size of the allocated memory
 */
void RealSpeak::text2Stream(const char *text, unsigned char **outbuffer, unsigned long *size)
{
    EnterCriticalSection(&criticalSection);
    strncpy(this->text, text, TTS_TEXT_BUFFER);

    // start conversion
    outputSize = 0;
    int ret = TtsProcess(hInst);
    if (ret != 0) {
	log->logf("<error> TTS: could not process input, got error: %d", ret);
	*size = 0;
	return;
    }

    log->logf("<debug> TTS: speech complete, generated %d bytes of data.", outputSize);

    
    *outbuffer = (unsigned char*)malloc (outputSize + 44);
    unsigned int wavelen = outputSize + 36;
    memcpy(&output[4], &wavelen, 4);
    wavelen = outputSize;
    memcpy(&output[40], &wavelen, 4);
    memcpy(*outbuffer, output, outputSize + 44);
    *size = outputSize + 44;
    
/*    FILE *fp = fopen("d:\\temp.wav", "wb");
    fwrite(*outbuffer, sizeof(unsigned char), outputSize + 44, fp);
    fclose(fp); */
    LeaveCriticalSection(&criticalSection);
}
